/*
 * HTML-object for the XForms library --- initial release, 0.1
 *
 * Copyright (C) 1996 Oliver Jakobs
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * The author can be contacted via email at O.Jakobs@bham.ac.uk
 * or at http://www-clg.bham.ac.uk/oliver
 *
 */

#include<ctype.h>
#include<stdio.h>
#include<stdlib.h>
#include "html.h"
#include "html_t.h"


/*
 * formatting commands
 */

#define LEFT	0
#define BLOCK	1

#define NONE		0
#define HEADING1	1
#define HEADING2	2
#define HEADING3	3
#define HEADING4	4
#define BOLD		5
#define ITALIC		6
#define LINK		7
#define PARAGRAPH	8
#define LINEBREAK	9
#define BLOCKQUOTE	10
#define LISTITEM	11
#define NUMLIST		12
#define BULLIST		13
#define HLINE		14
#define BULLET		15
#define CENTER		16
#define TITLE		17
#define TYPEWRITER	18


static CommandT htmlcommands[] = {
	{"h1", HEADING1, 2	},
	{"h2", HEADING2, 1	},
	{"h3", HEADING3, 1	},
	{"h4", HEADING4, 1	},
	{"b", BOLD,0	},
	{"strong", BOLD,0	},
	{"em", ITALIC,0	},
	{"tt", TYPEWRITER,0	},
	{"i", ITALIC,0	},
	{"p", PARAGRAPH, 2	},
	{"a", LINK, 0	},
	{"br", LINEBREAK,1	},
	{"ol", NUMLIST,0	},
	{"ul", BULLIST,0	},
	{"li", LISTITEM,2	},
	{"hr", HLINE,1	},
	{"title", TITLE,0	},
	{"center", CENTER,1	},
	{"blockquote", BLOCKQUOTE,1	},
	{ NULL, NONE,0 }
};


static WordT *
mkword(
	int x,
	int y,
	int col,
	int size,
	int style,
	char *word,
	char *link
)
{
	int asc, desc;
	WordT *tmp;

	tmp = (WordT*)malloc(sizeof(WordT));
	if(!tmp) return(NULL);
	tmp->x = x;
	tmp->y = y;
	tmp->word = NULL;
	if(word) {
		tmp->w = fl_get_string_width(style,size,word,strlen(word));
		tmp->h = fl_get_string_height(style,size,word,strlen(word),&asc,&desc);
		tmp->word = strdup(word);
	}
	tmp->col = col;
	tmp->size = size;
	tmp->style = style;
	if(link) tmp->link = strdup(link);
	else tmp->link = NULL;
	tmp->next = NULL;
	return(tmp);
}

static WordT *
addword(
	WordT *anchor,
	int x,
	int y,
	char *word,
	char *link,
	InfoT *info
)
{
	WordT *tmp;
	WordT *ptr = anchor;

	tmp = mkword(x,y,info->colour,info->fsize,info->fstyle,word,link);
	#ifdef DEBUG
		fprintf(stderr,"adding `%s'\n",word);
	#endif
	while(ptr && ptr->next) ptr = ptr->next;
	if(ptr) {
		ptr->next = tmp;
		return(anchor);
	} else {
		return(tmp);
	}
}

static WordT *
addspecial(
	WordT *anchor,
	int x,
	int y,
	int w,
	int h,
	char *item,
	InfoT *info
)
{
	WordT *tmp;
	WordT *ptr = anchor;
	tmp = mkword(x,y,info->colour,info->fsize,info->fstyle,NULL,item);
	tmp->w = w;
	tmp->h = h;
	while(ptr && ptr->next) ptr = ptr->next;
	if(ptr) {
		ptr->next = tmp;
		return(anchor);
	} else {
		return(tmp);
	}
}

static void
delword(WordT *word)
{
	if(!word) return;
	if(word->next) delword(word->next);
	free(word->word);
	free(word->link);
	free(word);
}

static void
delanchor(AnchorT *anchor)
{
	if(!anchor) return;
	if(anchor->next) delanchor(anchor->next);
	free(anchor);
}

static WordT *
addline(
	WordT *current,
	WordT *newline,
	InfoT *info
)
{
	WordT *tmp = current;
	WordT *ptr = newline;
	int width = 0;

	while(ptr) {
		width = ptr->x+ptr->w;
		ptr = ptr->next;
	}
	if(info->align == CENTER) {
		int diff = (info->width - width) /2;
		ptr = newline;
		while(ptr) {
			ptr->x += diff;
			ptr = ptr->next;
		}
	}
	
	while(tmp && tmp->next) tmp = tmp->next;
	if(tmp) {
		tmp->next = newline;
		return(current);
	} else {
		return(newline);
	}
}

static int
maxheight(
	InfoT *info,
	WordT *line
)
{
	int asc, desc;
	int height = fl_get_char_height(info->fstyle,info->fsize,&asc,&desc);
	while(line) {
		if(height < line->h) height = line->h;
		line = line->next;
	}
	return(height);
}


static InfoT *
mkinfo(
	int indent,
	int width,
	int colour,
	int fsize,
	int fstyle
)
{
	InfoT *tmp;
	tmp = (InfoT*)malloc(sizeof(InfoT));
	if(!tmp) return(NULL);
	tmp->x = 0;
	tmp->y = 0;
	tmp->align = LEFT;
	tmp->indent = indent;
	tmp->width = width;
	tmp->colour = colour;
	tmp->fsize = fsize;
	tmp->fstyle = fstyle;
	tmp->env = NONE;
	tmp->counter = 0;
	tmp->next = NULL;
	return(tmp);
}

static InfoT *
infodup(InfoT *info)
{
	InfoT *tmp;
	tmp = (InfoT*)malloc(sizeof(InfoT));
	if(!tmp) return(NULL);
	tmp->x = info->x;
	tmp->y = info->y;
	tmp->indent = info->indent;
	tmp->width = info->width;
	tmp->colour = info->colour;
	tmp->fsize = info->fsize;
	tmp->fstyle = info->fstyle;
	tmp->align = info->align;
	tmp->env = info->env;
	tmp->counter = info->counter;
	tmp->next = NULL;
	return(tmp);
}

static void
pushinfo(HtmlspecT *spec, InfoT *info)
{
	InfoT *tmp = infodup(info);
	#ifdef DEBUG
		printf("push\n");
	#endif
	tmp->next = spec->stack;
	spec->stack = tmp;
	return;
}

static InfoT *
popinfo(HtmlspecT *spec)
{
	InfoT *tmp = spec->stack;
	#ifdef DEBUG
		printf("pop\n");
	#endif
	if(!spec->stack) {
		return(NULL);
	}
	spec->stack = spec->stack->next;
	tmp->next = NULL;
	return(tmp);
}


static void
delinfo(InfoT *info)
{
	if(!info) return;
	if(info->next) delinfo(info->next);
	free(info);
}


static void
parsecommand(
	char *markup,
	int *col,
	int *command,
	int *begin,
	int *lines,
	char ***attribs
)
{
	char buffer[80];
	static char attrib[200];
	int i;
	*command = NONE;
	*begin = 1;
	*attribs = NULL;

	if(markup[(*col)] == '\0' || markup[(*col)] != '<') return;
	(*col)++;
	if(markup[(*col)] == '/') {
		*begin = 0;
		(*col)++;
	}
	for(i = 0; markup[(*col)] && !isspace(markup[(*col)]); i++) {
		buffer[i] = tolower(markup[(*col)++]);
		buffer[i+1] = '\0';
		if(buffer[i] == '>') {
			buffer[i] = '\0';
			(*col)--;
			break;
		}
	}
	#ifdef DEBUG
		fprintf(stderr,"Command: `%s' [%d]\n",buffer,*begin);
	#endif
	for(i = 0; htmlcommands[i].name; i++) {
		if(!strcmp(htmlcommands[i].name,buffer)) {
			*command = htmlcommands[i].command;
			*lines = htmlcommands[i].linebreaks;
			break;
		}
	}
	
	while(markup[(*col)] && markup[(*col)] == ' ') (*col)++;
	i = 0;
	while(markup[(*col)] && markup[(*col)] != '>') {
		attrib[i++] = markup[(*col)++];
	}
	attrib[i] = '\0';
	if(markup[(*col)] == '>') (*col)++;
	if(strlen(attrib)) {
		int a = 0;
		for(i = 0; attrib[i]; i++) if(attrib[i] == '=') a++;
		*attribs = (char**)malloc((a+1) * sizeof(char*));
		(*attribs)[a] = NULL;
		i = 0;
		while(a) {
			while(attrib[i] && isspace(attrib[i])) i++;
			(*attribs)[a-1] = attrib+i;
			while(attrib[i] && !isspace(attrib[i])) i++;
			if(attrib[i] == ' ') attrib[i++] = '\0';
			a--;
		}
	}
	return;
}

void
fl_html_redraw(FL_OBJECT *ob)
{
	if(!ob || ob->objclass != FL_HTML) return;
	else {
		HtmlspecT *spec = (HtmlspecT*)ob->spec;
		WordT *tmp = spec->data;

		#ifdef DEBUG
			fprintf(stderr,"<in redraw %d>\n",spec->topline);
		#endif

		fl_rectf(ob->x,ob->y,ob->w,ob->h,ob->col1);
		fl_set_clipping(ob->x,ob->y,ob->w,ob->h);
		while(tmp && tmp->y < spec->topline) {
			#ifdef DEBUG
				fprintf(stderr,"skipping %s (y=%d)\n",tmp->word,tmp->y);
			#endif
			tmp = tmp->next;
		}
		while(tmp && tmp->y+tmp->h-spec->topline < ob->h) {
			if(!tmp->word && tmp->link) switch(tmp->link[0]) {
			case 'l':	/* line */
				fl_line(ob->x+tmp->x,ob->y+tmp->y-spec->topline,
					tmp->w,1,tmp->col);
				break;
			case 'b':	/* bullet */
				{
					int x = ob->x+tmp->x + tmp->w/3;
					int y = ob->y+tmp->y-spec->topline + tmp->h/3;
					fl_ovalf(x,y,tmp->w*2/3,tmp->w*2/3,tmp->col);
				}
				break;
			} else {
				fl_drw_text_cursor(FL_ALIGN_CENTER,ob->x+tmp->x,
					ob->y+tmp->y-spec->topline,
					tmp->w,tmp->h,tmp->col,tmp->size,tmp->style,
					tmp->word,0,-1);
			}
			tmp = tmp->next;
		}
	}
	fl_unset_clipping();
}




static int
handle_html(
	FL_OBJECT *ob,
	int event,
	FL_Coord mx,
	FL_Coord my,
	int key,
	void *xev
)
{
	HtmlspecT *spec = (HtmlspecT*)ob->spec;
	WordT *word = spec->data;
	/*ButtonT *button = spec->buttons;*/
	switch(event) {
		case FL_DRAW:
			fl_html_redraw(ob);
			return(0);
		case FL_FREEMEM:
			delanchor(spec->anchors);
			/*delbutton(spec->buttons);*/
			delword(spec->data);
			fl_free(((HtmlspecT*)ob->spec));
			return(0);
		case FL_RELEASE:
			free(spec->lastword); spec->lastword = NULL;
			free(spec->lastlink); spec->lastlink = NULL;
			if(ob->type != FL_PASSIVE_HTML) {
				mx -= ob->x;
				my -= ob->y;
				my += spec->topline;
				while(word) {
					if(word->word && mx >= word->x && mx <= word->x+word->w
					&& my >= word->y && my <= word->y+word->h) {
						spec->lastword = strdup(word->word);
						if(word->link) spec->lastlink= strdup(word->link);
						else spec->lastlink = NULL;
						#ifdef DEBUG
							fprintf(stderr,"Word: '%s', link: '%s'\n",
								spec->lastword,spec->lastlink);
						#endif
						return(1);
						break;
					}
					word = word->next;
				}
			}
			break;
		default:
			return(0);
	}
	return(0);
}

static int
font(
	int current,
	int bold,
	int italic
)
{
	return((4*(current / 4)) + (bold + 2*italic));
}

void
fl_html_set_data(
	FL_OBJECT *ob,
	WordT *data
)
{
	HtmlspecT *spec = (HtmlspecT*)ob->spec;
	if(!ob || ob->objclass != FL_HTML) return;
	delword(spec->data);
	spec->data = data;
}
	

static WordT *
format(
	FL_OBJECT *ob,
	char **text,
	int width
)
{
	WordT *data = NULL;
	WordT *currentline = NULL;
	InfoT *info = NULL;
	char *url = NULL;
	int line,col;
	int x,y;
	HtmlspecT *spec;
	const char *delim = "\n\t <";
	char titlebuf[100];

	if(!ob || ob->objclass != FL_HTML || !text) return(NULL);

	spec = (HtmlspecT*) ob->spec;
	if(!spec) return(NULL);

	info = spec->p_info;
	currentline = spec->p_currentline;
	url = spec->p_url;
	x = spec->p_x;
	y = spec->p_y;
	data = spec->data;
	
	if(!info) {	/* initial data */
		if(width == 0) width = (ob->w *38) /40;

		/*
	 	 * set default env
	 	 */
		info = mkinfo(ob->w/40 ,width,ob->col2,spec->fontsize,
					spec->fontstyle);

		x = 0; y = ob->h/40;
		info->x = x;
		info->y = y;
		info->align = LEFT;
	}

	line = col = 0;

	/*
	 * read next token
	 */
	do {
		char buffer[80];
		int command;
		int begin;
		int linebreaks;
		char **attr = NULL;

		if(!info) break;	/* mark-up imbalance */
		while(text[line] && text[line][col] == '\0') {
			line++;
			col = 0;
		}
		if(!text[line]) break;	/* finished */
		while(text[line][col] && isspace(text[line][col])) col++;
		if(text[line][col] && text[line][col] != '<') {	/* word */
			int i = 0;
			while(text[line][col] && isspace(text[line][col])) col++;
			if(text[line][col] == '\0') continue;
			while(text[line][col] && !strchr(delim,text[line][col])) {
				buffer[i++] = text[line][col++];
			}
			buffer[i] = '\0'; 
			if(info->env == TITLE) {
				strcat(titlebuf,buffer);
				strcat(titlebuf," ");
				continue;
			}
			width = fl_get_string_width(info->fstyle, info->fsize,
						buffer, strlen(buffer));
			if(x+width < info->width) {	/* word fits */
				currentline = addword(currentline,x,y,buffer,url,info);
				x = x + width + fl_get_string_width(info->fstyle, info->fsize,
						"_",1);
			} else {
				data = addline(data,currentline,info);
				y += maxheight(info,currentline);
				x = info->indent;
				currentline = addword(NULL,x,y,buffer,url,info);
				x = x + width + fl_get_string_width(info->fstyle, info->fsize,
						"_",1);
			}
		} 
		if(text[line][col] && text[line][col] == '<') {
			if(attr) free(attr);
			parsecommand(text[line],&col, &command, &begin, &linebreaks, &attr);
			switch(command) {
			case HEADING1: case HEADING2: case HEADING3: case HEADING4:
			case LISTITEM: case CENTER:
				if(data) {
					y += maxheight(info,currentline);
					data = addline(data,currentline,info);
					currentline = NULL;
				}
				x = info->indent;
				break;
			default:
				break;
			}
			linebreaks = 0;
			switch(command) {	/* command */
		case TITLE:
			if(begin) {
				pushinfo(spec,info);
				info->env = TITLE;
				titlebuf[0] = '\0';
			} else {
				delinfo(info);
				info = popinfo(spec);
				/* change title of form window here... */
			}
		case HEADING1:
			if(begin) {
				pushinfo(spec,info);
				info->fsize = FL_LARGE_SIZE;
				info->fstyle= font(info->fstyle,1,0);
				if(data) linebreaks = 1;
			} else {
				delinfo(info);
				info = popinfo(spec);
				linebreaks = 1;
			}
			break;
		case HEADING2:
			if(begin) {
				pushinfo(spec,info);
				info->fsize = FL_MEDIUM_SIZE;
				info->fstyle= font(info->fstyle,1,0);
				if(data) linebreaks = 1;
			} else {
				delinfo(info);
				info = popinfo(spec);
				linebreaks = 1;
			}
			break;
		case HEADING3:
			if(begin) {
				pushinfo(spec,info);
				info->fsize = spec->fontsize;
				info->fstyle= font(info->fstyle,1,0);
				if(data) linebreaks = 1;
			} else {
				delinfo(info);
				info = popinfo(spec);
				linebreaks = 1;
			}
			break;
		case HEADING4:
			if(begin) {
				pushinfo(spec,info);
				info->fsize = spec->fontsize;
				info->fstyle= font(info->fstyle,0,1);
				if(data) linebreaks = 1;
			} else {
				delinfo(info);
				info = popinfo(spec);
				linebreaks = 1;
			}
			break;
		case BOLD:
			if(begin) {
				pushinfo(spec,info);
				info->fstyle= font(info->fstyle,1,0);
			} else {
				delinfo(info);
				info = popinfo(spec);
			}
			break;
		case TYPEWRITER:
			if(begin) {
				pushinfo(spec,info);
				info->fstyle= FL_FIXED_STYLE;
			} else {
				delinfo(info);
				info = popinfo(spec);
			}
			break;
		case ITALIC:
			if(begin) {
				pushinfo(spec,info);
				info->fstyle= font(info->fstyle,0,1);
			} else {
				delinfo(info);
				info = popinfo(spec);
			}
			break;
		case BLOCKQUOTE:
			if(begin) {
				pushinfo(spec,info);
				info->indent += info->width/20;
				info->width = (info->width * 18)/20;
				x = info->indent;
				linebreaks = 1;
			} else {
				delinfo(info);
				info = popinfo(spec);
				linebreaks = 1;
			}
			break;
		case CENTER:
			if(begin) {
				pushinfo(spec,info);
				info->align = CENTER;
			} else {
				delinfo(info);
				info = popinfo(spec);
				x = info->indent;
			}
			break;
		case PARAGRAPH:
			linebreaks = 2;
			break;
		case LINEBREAK:
			linebreaks = 1;
			break;
		case LISTITEM:
			if(currentline) {
				data = addline(data,currentline,info);
				y += maxheight(info,currentline);
				currentline = NULL;
			}
			x = info->indent;
			switch(info->env) {
			case NUMLIST: {
				char buffer[40];
				sprintf(buffer,"%d. ",++info->counter);
				currentline = addword(currentline,
					x - fl_get_string_width(info->fstyle,info->fsize,
						buffer,strlen(buffer)), y,buffer,url,info);
				break;
				}
			case BULLIST: {
				int asc, desc;
				int width = fl_get_string_width(info->fstyle, info->fsize,
								"_",1);
				currentline = addspecial(currentline,x-width*3/2,y, width,
					fl_get_string_height(info->fstyle, info->fsize,"_",1,
						&asc,&desc),"b",info);
				break;
				}
			default:
				break;
			}
			break;
		case NUMLIST:
			if(begin) {
				int width = fl_get_string_width(
					info->fstyle,
					info->fsize,"99. ",strlen("99. "));
				pushinfo(spec,info);
				info->counter = 0;
				info->env = NUMLIST;
				info->indent += width;
				info->width -= (2*width);
				linebreaks = 1;
			} else {
				delinfo(info);
				info = popinfo(spec);
				x = info->indent;
				linebreaks = 1;
			}
			break;
		case BULLIST:
			if(begin) {
				int width = fl_get_string_width(
					info->fstyle,
					info->fsize,"* ",strlen("* "));
				int outdent = 2*width;
				pushinfo(spec,info);
				info->counter = 0;
				info->env = BULLIST;
				info->indent += outdent;
				info->width -= (width + outdent);
				linebreaks = 1;
			} else {
				delinfo(info);
				info = popinfo(spec);
				x = info->indent;
				linebreaks = 1;
			}
			break;
		case LINK:
			if(begin) {
				int i,j;
				if(attr) for(i = 0; attr[i]; i++) {
					#ifdef DEBUG
						fprintf(stderr,"Attr[%d]: `%s'\n",i,attr[i]);
					#endif
					for(j = 0; attr[i][j] && attr[i][j] != '='; j++)
						attr[i][j] = tolower(attr[i][j]);
					if(!strncmp(attr[i],"href",strlen("href"))) {
						char *tmp = strchr(attr[i],'=');
						if(!tmp) continue;
						tmp++;
						if(url) free(url);
						url = strdup(tmp);
						pushinfo(spec,info);
						info->colour = spec->highlight;
						#ifdef DEBUG
							fprintf(stderr,"URL:`%s'\n",url);
						#endif
					}
					if(!strncmp(attr[i],"name",strlen("name"))) {
						AnchorT *anch = (AnchorT*)malloc(sizeof(AnchorT));
						char *tmp = strchr(attr[i],'=');
						if(!anch || !tmp) continue;
						tmp++;
						anch->name = strdup(tmp);
						anch->yval = y;
						anch->next = spec->anchors;
						spec->anchors = anch;
					}
				}
			} else {
				delinfo(info);
				info = popinfo(spec);
				if(url) free(url);
				url = NULL;
			}
			break;
		case HLINE: {
				int asc, desc;
				int height = fl_get_string_height(info->fstyle, info->fsize,
								"_",1,&asc,&desc);
				currentline = addspecial(currentline,x,y+height/2,
					info->width - x, height,"l",info);
				linebreaks = 1;
			}
			break;
		case BULLET: {
				int asc, desc;
				currentline = addspecial(currentline,x,y,
					fl_get_string_width(info->fstyle, info->fsize,"_",1),
					fl_get_string_height(info->fstyle, info->fsize,"_",1,
						&asc,&desc),"b",info);
				x = info->width;
			}
			break;
		default:
			#ifdef DEBUG
				fprintf(stderr,"Unknown command\n");
			#endif
			break;
		}	/* switch */
			while(linebreaks--) {
				x = info->indent;
				y += maxheight(info,currentline);
				data = addline(data,currentline,info);
				currentline = NULL;
			}
		}
	} while(1 == 1);
	spec->height = y+1;
	spec->p_info = info;
	spec->p_currentline = currentline;
	spec->p_url = url;
	spec->p_x = x;
	spec->p_y = y;
	return(data);
}

static void
scrollbar_cb(FL_OBJECT *ob, long data)
{
	FL_OBJECT *html = (FL_OBJECT*)data;
	double x = fl_get_slider_value(ob);
	if(html) fl_html_jmp(html,(int)-x);
}

static void
update_sbar(FL_OBJECT *ob)
{
	HtmlspecT *spec;
	spec = (HtmlspecT*) ob->spec;
	if(spec->sbar) {
		fl_set_slider_value(spec->sbar,(double)(spec->topline)/100.0);
		fl_set_slider_size(spec->sbar,((double)spec->height/(double)ob->h)/100);
	}
}


int
fl_html_set_text(
	FL_OBJECT *ob,
	char **text,
	int width
)
{
	WordT *data = format(ob,text,width);
	fl_html_set_data(ob,data);
	return(1);
}

int
fl_html_add_line(
	FL_OBJECT *ob,
	char *line
)
{
	char *list[2];
	HtmlspecT *spec;

	list[0] = line;
	list[1] = NULL;
	spec = (HtmlspecT*) ob->spec;
	spec->data = format(ob,list,0);
	update_sbar(ob);
	return(1);
}

const char *
fl_html_get_word(
	FL_OBJECT *ob
)
{
	HtmlspecT *spec;
	spec = (HtmlspecT*) ob->spec;
	return(spec->lastword);
}

const char *
fl_html_get_link(
	FL_OBJECT *ob
)
{
	HtmlspecT *spec;
	spec = (HtmlspecT*) ob->spec;
	return(spec->lastlink);
}

int
fl_html_set_scrollbar(
	FL_OBJECT *ob,
	FL_OBJECT *sbar
)
{
	HtmlspecT *spec;
	spec = (HtmlspecT*) ob->spec;
	spec->sbar = sbar;
	fl_set_slider_value(sbar,(double)(spec->topline)/100.0);
	fl_set_slider_bounds(sbar,(double)0.0,(double)100.0);
	fl_set_slider_size(sbar,((double)spec->height/(double)ob->h)/100);
	fl_set_object_callback(sbar,scrollbar_cb,(long)ob);
	/*fl_redraw_object(sbar);*/
	return(1);
}

/*
 * scroll
 */

void fl_html_pg(
	FL_OBJECT *ob,
	int dir
)
{
	HtmlspecT *spec;
	spec = (HtmlspecT*) ob->spec;
	if(dir < 0) {
		spec->topline -= ob->h;
		if(spec->topline < 0) spec->topline = 0;
	} else if(dir > 0) {
		spec->topline += ob->h;
		if(spec->topline > spec->height) spec->topline = spec->height - ob->h;
	}
	update_sbar(ob);
	fl_html_redraw(ob);
}

/*
 * scroll one line
 */

void fl_html_scroll(
	FL_OBJECT *ob,
	int dir
)
{
	HtmlspecT *spec;
	int asc,desc;
	int delta;

	spec = (HtmlspecT*) ob->spec;
	delta = fl_get_string_height(spec->fontstyle,spec->fontsize,"X",1,
				&asc,&desc);
	
	if(dir < 0) {
		spec->topline -= delta;
		if(spec->topline < 0) spec->topline = 0;
	} else if(dir > 0) {
		spec->topline += delta;
		if(spec->topline > spec->height) spec->topline = spec->height - ob->h;
	}
	update_sbar(ob);
	fl_html_redraw(ob);
}

int
fl_html_jmp(
	FL_OBJECT *ob,
	int percentage
)
{
	HtmlspecT *spec;
	
	if(abs(percentage) > 100) return(-1);
	spec = (HtmlspecT*) ob->spec;
	spec->topline = spec->height * abs(percentage) / 100;
	if(percentage > 0) update_sbar(ob);
	fl_html_redraw(ob);
	return(1);
}

int
fl_html_jmp_anchor(
	FL_OBJECT *ob,
	const char *name
)
{
	HtmlspecT *spec;
	AnchorT *anchor;
	spec = (HtmlspecT*) ob->spec;
	anchor = spec->anchors;
	while(anchor) {
		if(anchor->name && !strcmp(anchor->name,name)) {
			spec->topline = anchor->yval;
			update_sbar(ob);
			fl_html_redraw(ob);
			return(1);
		} else anchor = anchor->next;
	}
	return(0);
}

/*
 * create an html object
 */

FL_OBJECT *
fl_create_html(
	int type,
	FL_Coord x,
	FL_Coord y,
	FL_Coord w,
	FL_Coord h,
	const char *label)
{
	FL_OBJECT *ob;
	HtmlspecT *spec;

	/*
	 * create generic object
	 */
	ob = fl_make_object(FL_HTML,type,x,y,w,h,label,handle_html);

	ob->col1 = FL_WHITE;	/* background */
	ob->col2 = FL_BLACK;	/* foreground */
	ob->wantkey = FL_KEY_SPECIAL;
	
	spec = (HtmlspecT*) ob->spec = fl_calloc(1, sizeof(HtmlspecT));
	spec->height = 0;
	spec->fontsize = FL_MEDIUM_SIZE;
	spec->fontstyle = FL_NORMAL_STYLE;
	spec->topline = 0;
	spec->anchors = NULL;
	spec->data = NULL;
	spec->lastword = NULL;
	spec->lastlink = NULL;
	spec->highlight = FL_BLUE;
	spec->sbar = NULL;
	spec->stack = NULL;
	spec->p_info = NULL;
	spec->p_currentline = NULL;
	spec->p_url = NULL;
	spec->p_x = 0;
	spec->p_y = 0;

	return(ob);
}

/*
 * add object to a form
 */

FL_OBJECT *
fl_add_html(
	int type,
	FL_Coord x,
	FL_Coord y,
	FL_Coord w,
	FL_Coord h,
	const char *label)
{
	FL_OBJECT *ob = fl_create_html(type,x,y,w,h,label);
	fl_add_object(fl_current_form,ob);
	return(ob);
}
